@@ -26,7 +26,7 @@ extern "C" {
 
 #define TICKIT_VERSION_MAJOR 0
 #define TICKIT_VERSION_MINOR 4
-#define TICKIT_VERSION_PATCH 0
+#define TICKIT_VERSION_PATCH 2
 
 int tickit_version_major(void);
 int tickit_version_minor(void);
@@ -126,6 +126,11 @@ typedef enum {
   TICKIT_N_PEN_ATTRS
 } TickitPenAttr;
 
+typedef struct {
+  pid_t pid;
+  int wstatus;
+} TickitProcessWatchInfo;
+
 /* additional attribute types recognised by tickit_pen_new_attrs */
 enum {
   /* We're unlikely to ever have 256 attributes, so adding 0x100 should be safe */
@@ -442,6 +447,8 @@ struct TickitTermBuilder {
 };
 TickitTerm *tickit_term_build(const struct TickitTermBuilder *builder);
 
+void tickit_term_teardown(TickitTerm *tt);
+
 TickitTerm *tickit_term_ref(TickitTerm *tt);
 void        tickit_term_unref(TickitTerm *tt);
 
@@ -628,7 +635,7 @@ void tickit_renderbuffer_moverect(TickitRenderBuffer *rb, const TickitRect *dest
 
 void tickit_renderbuffer_flush_to_term(TickitRenderBuffer *rb, TickitTerm *tt);
 
-void tickit_renderbuffer_blit(TickitRenderBuffer *dst, TickitRenderBuffer *src);
+void tickit_renderbuffer_blit(TickitRenderBuffer *dst, const TickitRenderBuffer *src);
 
 // This API is still somewhat experimental
 
@@ -797,6 +804,10 @@ void *tickit_watch_timer_at_tv(Tickit *t, const struct timeval *at, TickitBindFl
 
 void *tickit_watch_later(Tickit *t, TickitBindFlags flags, TickitCallbackFn *fn, void *user);
 
+void *tickit_watch_signal(Tickit *t, int signum, TickitBindFlags flags, TickitCallbackFn *fn, void *user);
+
+void *tickit_watch_process(Tickit *t, pid_t pid, TickitBindFlags flags, TickitCallbackFn *fn, void *user);
+
 void tickit_watch_cancel(Tickit *t, void *watch);
 
 /* Debug support */
