@@ -5,8 +5,16 @@
 #include <signal.h>
 #include <string.h>
 #include <sys/time.h>
+#include <sys/wait.h>
 #include <unistd.h>
 
+/* POSIX doesn't actually define an NSIG macro, but if it did, almost every
+ * known platform would set it to 32
+ */
+#ifndef NSIG
+#  define NSIG 32
+#endif
+
 #define streq(a,b) (!strcmp(a,b))
 
 /* INTERNAL */
@@ -22,6 +30,8 @@ struct TickitWatch {
     WATCH_IO,
     WATCH_TIMER,
     WATCH_LATER,
+    WATCH_SIGNAL,
+    WATCH_PROCESS,
   } type;
 
   TickitBindFlags flags;
@@ -42,6 +52,15 @@ struct TickitWatch {
     struct {
       struct timeval at;
     } timer;
+
+    struct {
+      int signum;
+    } signal;
+
+    struct {
+      pid_t pid;
+      int wstatus; /* in case of pre-exited process */
+    } process;
   };
 };
 
@@ -53,11 +72,20 @@ struct Tickit {
   TickitTerm   *term;
   TickitWindow *rootwin;
 
-  TickitWatch *iowatches, *timers, *laters;
+  TickitWatch *iowatches, *timers, *laters, *signals, *processes;
 
   const TickitEventHooks *evhooks;
   void                   *evdata;
 
+  struct {
+    int pipefds[2];
+    void *pipewatch;
+    sigset_t watched;
+    sigset_t pending;
+  } signal;
+
+  void *sigchldwatch;
+
   unsigned int done_setup    : 1,
                use_altscreen : 1;
 };
@@ -80,6 +108,50 @@ static int on_term_readable(Tickit *t, TickitEventFlags flags, void *info, void
   return 0;
 }
 
+static Tickit *signal_observer;
+
+static void sighandler(int signum)
+{
+  if(signal_observer) {
+    sigaddset(&signal_observer->signal.pending, signum);
+    write(signal_observer->signal.pipefds[1], "\0", 1);
+  }
+}
+
+static int on_sigpipe_readable(Tickit *t, TickitEventFlags flags, void *info, void *user)
+{
+  char buf[1];
+  read(t->signal.pipefds[0], &buf, 1);
+
+  sigset_t pending;
+  {
+    sigset_t orig;
+    sigprocmask(SIG_BLOCK, &t->signal.watched, &orig);
+
+    pending = t->signal.pending;
+    sigemptyset(&t->signal.pending);
+
+    sigprocmask(SIG_SETMASK, &orig, NULL);
+  }
+
+  TickitWatch *this;
+  for(this = t->signals; this; this = this->next) {
+    if(sigismember(&pending, this->signal.signum))
+      (*this->fn)(this->t, TICKIT_EV_FIRE, NULL, this->user);
+  }
+
+  return 0;
+}
+
+static int on_sigwinch(Tickit *t, TickitEventFlags flags, void *info, void *user)
+{
+  if(!t->term)
+    return 0;
+
+  tickit_term_refresh_size(t->term);
+  return 0;
+}
+
 static void setupterm(Tickit *t)
 {
   TickitTerm *tt = t->term;
@@ -124,6 +196,15 @@ Tickit *tickit_build(const struct TickitBuilder *builder)
   t->iowatches = NULL;
   t->timers    = NULL;
   t->laters    = NULL;
+  t->signals   = NULL;
+  t->processes = NULL;
+
+  t->signal.pipefds[0] = -1;
+  t->signal.pipewatch = NULL;
+  sigemptyset(&t->signal.watched);
+  sigemptyset(&t->signal.pending);
+
+  t->sigchldwatch = NULL;
 
   t->done_setup = false;
 
@@ -144,6 +225,8 @@ Tickit *tickit_build(const struct TickitBuilder *builder)
   tickit_watch_io(t, tickit_term_get_input_fd(tt), TICKIT_IO_IN,
       0, on_term_readable, NULL);
 
+  tickit_watch_signal(t, SIGWINCH, 0, on_sigwinch, NULL);
+
   return t;
 
 abort:
@@ -184,6 +267,17 @@ Tickit *tickit_new_stdtty(void)
   });
 }
 
+static void insert_watch(TickitWatch **watchesptr, TickitBindFlags flags, TickitWatch *new)
+{
+  if(!(flags & TICKIT_BIND_FIRST)) {
+    while(*watchesptr)
+      watchesptr = &(*watchesptr)->next;
+  }
+
+  new->next = *watchesptr;
+  *watchesptr = new;
+}
+
 static void destroy_watchlist(Tickit *t, TickitWatch *watches, void (*cancelfunc)(void *data, TickitWatch *watch))
 {
   TickitWatch *this, *next;
@@ -200,6 +294,44 @@ static void destroy_watchlist(Tickit *t, TickitWatch *watches, void (*cancelfunc
   }
 }
 
+static void invoke_watch(TickitWatch *watch, TickitEventFlags flags, void *info)
+{
+  (*watch->fn)(watch->t, flags, info, watch->user);
+
+  /* Remove oneshot watches from the list */
+  TickitWatch **prevp;
+  switch(watch->type) {
+    case WATCH_NONE:
+    case WATCH_IO:
+    case WATCH_SIGNAL:
+      return;
+
+    case WATCH_TIMER:
+      prevp = &watch->t->timers;
+      break;
+
+    case WATCH_LATER:
+      prevp = &watch->t->laters;
+      break;
+
+    case WATCH_PROCESS:
+      prevp = &watch->t->processes;
+      break;
+  }
+
+  while(*prevp) {
+    if(*prevp == watch) {
+      *prevp = watch->next;
+      watch->next = NULL;
+      watch->type = WATCH_NONE;
+      free(watch);
+      return;
+    }
+
+    prevp = &(*prevp)->next;
+  }
+}
+
 static void tickit_destroy(Tickit *t)
 {
   if(t->done_setup)
@@ -212,12 +344,28 @@ static void tickit_destroy(Tickit *t)
     tickit_term_unref(t->term);
   }
 
+  if(t->sigchldwatch)
+    tickit_watch_cancel(t, t->sigchldwatch);
+
+  if(t->signal.pipewatch)
+    tickit_watch_cancel(t, t->signal.pipewatch);
+  for(int signum = 1; signum < NSIG; signum++) {
+    if(sigismember(&t->signal.watched, signum))
+      sigaction(signum, &(struct sigaction){ .sa_handler = SIG_DFL }, NULL);
+  }
+  if(signal_observer == t)
+    signal_observer = NULL;
+
   if(t->iowatches)
     destroy_watchlist(t, t->iowatches, t->evhooks->cancel_io);
   if(t->timers)
     destroy_watchlist(t, t->timers, t->evhooks->cancel_timer);
   if(t->laters)
     destroy_watchlist(t, t->laters, t->evhooks->cancel_later);
+  if(t->signals)
+    destroy_watchlist(t, t->signals, t->evhooks->cancel_signal);
+  if(t->processes)
+    destroy_watchlist(t, t->processes, t->evhooks->cancel_process);
 
   (*t->evhooks->destroy)(t->evdata);
 
@@ -277,15 +425,10 @@ bool tickit_setctl_int(Tickit *t, TickitCtl ctl, int value)
   return false;
 }
 
-// TODO: copy the entire SIGWINCH-like structure from term.c
-// For now we only handle atmost-one running Tickit instance
-
-static Tickit *running_tickit;
-
-static void sigint(int sig)
+static int on_sigint(Tickit *t, TickitEventFlags flags, void *info, void *data)
 {
-  if(running_tickit)
-    tickit_stop(running_tickit);
+  tickit_stop(t);
+  return 0;
 }
 
 void tickit_tick(Tickit *t, TickitRunFlags flags)
@@ -298,15 +441,14 @@ void tickit_tick(Tickit *t, TickitRunFlags flags)
 
 void tickit_run(Tickit *t)
 {
-  running_tickit = t;
-  signal(SIGINT, sigint);
+  void *sigint_watch = tickit_watch_signal(t, SIGINT, 0, &on_sigint, NULL);
 
   if(!t->done_setup)
     setupterm(t);
 
   (*t->evhooks->run)(t->evdata, TICKIT_RUN_DEFAULT);
 
-  running_tickit = NULL;
+  tickit_watch_cancel(t, sigint_watch);
 }
 
 void tickit_stop(Tickit *t)
@@ -334,12 +476,7 @@ void *tickit_watch_io(Tickit *t, int fd, TickitIOCondition cond, TickitBindFlags
   if(!(*t->evhooks->io)(t->evdata, fd, cond, flags, watch))
     goto fail;
 
-  TickitWatch **prevp = &t->iowatches;
-  while(*prevp)
-    prevp = &(*prevp)->next;
-
-  watch->next = *prevp;
-  *prevp = watch;
+  insert_watch(&t->iowatches, flags, watch);
 
   return watch;
 
@@ -374,7 +511,7 @@ void *tickit_watch_timer_at_tv(Tickit *t, const struct timeval *at, TickitBindFl
       goto fail;
 
   TickitWatch **prevp = &t->timers;
-  /* Try to insert in-order at matching timestamp */
+  /* Try to insert in-order at matching timestamp; don't use insert_watch() */
   while(*prevp && !timercmp(&(*prevp)->timer.at, at, >))
     prevp = &(*prevp)->next;
 
@@ -433,12 +570,7 @@ void *tickit_watch_later(Tickit *t, TickitBindFlags flags, TickitCallbackFn *fn,
     if(!(*t->evhooks->later)(t->evdata, flags, watch))
       goto fail;
 
-  TickitWatch **prevp = &t->laters;
-  while(*prevp)
-    prevp = &(*prevp)->next;
-
-  watch->next = *prevp;
-  *prevp = watch;
+  insert_watch(&t->laters, flags, watch);
 
   return watch;
 
@@ -447,6 +579,125 @@ fail:
   return NULL;
 }
 
+static void watch_signal(Tickit *t, int signum, TickitWatch *watch)
+{
+  if(t->signal.pipefds[0] == -1) {
+    pipe(t->signal.pipefds);
+    t->signal.pipewatch = tickit_watch_io(t, t->signal.pipefds[0], TICKIT_IO_IN, 0, &on_sigpipe_readable, t);
+  }
+
+  if(sigismember(&t->signal.watched, signum))
+    return;
+
+  sigaction(signum, &(struct sigaction){ .sa_handler = sighandler }, NULL);
+  sigaddset(&t->signal.watched, signum);
+
+  if(!signal_observer)
+    signal_observer = t;
+}
+
+static void unwatch_signal(Tickit *t, TickitWatch *watch)
+{
+  int signum = watch->signal.signum;
+
+  /* watch has already been removed from t->signals */
+  TickitWatch *this;
+  for(this = t->signals; this; this = this->next) {
+    if(this->signal.signum == signum)
+      return;
+  }
+
+  sigdelset(&t->signal.watched, signum);
+  sigaction(signum, &(struct sigaction){ .sa_handler = SIG_DFL }, NULL);
+}
+
+void *tickit_watch_signal(Tickit *t, int signum, TickitBindFlags flags, TickitCallbackFn *fn, void *user)
+{
+  TickitWatch *watch = malloc(sizeof(TickitWatch));
+  if(!watch)
+    return NULL;
+
+  watch->next = NULL;
+  watch->t    = t;
+  watch->type = WATCH_SIGNAL;
+
+  watch->flags = flags & (TICKIT_BIND_UNBIND|TICKIT_BIND_DESTROY);
+  watch->fn = fn;
+  watch->user = user;
+
+  watch->signal.signum = signum;
+
+  if(!t->evhooks->signal ||
+      !(*t->evhooks->signal)(t->evdata, signum, flags, watch))
+    watch_signal(t, signum, watch);
+
+  insert_watch(&t->signals, flags, watch);
+
+  return watch;
+}
+
+static int on_sigchld(Tickit *t, TickitEventFlags flags, void *info, void *data)
+{
+  TickitWatch *this, *next;
+  for(this = t->processes; this; this = next) {
+    next = this->next;
+
+    TickitProcessWatchInfo info;
+    if(waitpid(this->process.pid, &info.wstatus, WNOHANG) <= 0)
+      continue;
+
+    info.pid = this->process.pid;
+    invoke_watch(this, TICKIT_EV_FIRE, &info);
+  }
+  return 0;
+}
+
+static int process_notify(Tickit *t, TickitEventFlags flags, void *_info, void *data)
+{
+  TickitWatch *watch = data;
+
+  tickit_evloop_invoke_processwatch(watch, TICKIT_EV_FIRE, watch->process.wstatus);
+
+  return 0;
+}
+
+void *tickit_watch_process(Tickit *t, pid_t pid, TickitBindFlags flags, TickitCallbackFn *fn, void *user)
+{
+  TickitWatch *watch = malloc(sizeof(TickitWatch));
+  if(!watch)
+    return NULL;
+
+  watch->next = NULL;
+  watch->t    = t;
+  watch->type = WATCH_PROCESS;
+
+  watch->flags = flags & (TICKIT_BIND_UNBIND|TICKIT_BIND_DESTROY);
+  watch->fn = fn;
+  watch->user = user;
+
+  watch->process.pid = pid;
+
+  if(!t->evhooks->process ||
+      !(*t->evhooks->process)(t->evdata, pid, flags, watch)) {
+    if(!t->sigchldwatch)
+      t->sigchldwatch = tickit_watch_signal(t, SIGCHLD, 0, &on_sigchld, NULL);
+
+    if(waitpid(pid, &watch->process.wstatus, WNOHANG) > 0) {
+      /* Process already exited, so SIGCHLD won't see it. We can't invoke
+       * callback immediately as user will be expecting it to only be called via
+       * tickit_run(). We'll install a later handler for it
+       */
+      tickit_watch_later(t, 0, process_notify, watch);
+
+      return watch;
+    }
+  }
+
+  insert_watch(&t->processes, flags, watch);
+
+  return watch;
+}
+
 void tickit_watch_cancel(Tickit *t, void *_watch)
 {
   TickitWatch *watch = _watch;
@@ -462,8 +713,14 @@ void tickit_watch_cancel(Tickit *t, void *_watch)
     case WATCH_LATER:
       thisp = &t->laters;
       break;
+    case WATCH_SIGNAL:
+      thisp = &t->signals;
+      break;
+    case WATCH_PROCESS:
+      thisp = &t->processes;
+      break;
 
-    default:
+    case WATCH_NONE:
       return;
   }
 
@@ -487,8 +744,18 @@ void tickit_watch_cancel(Tickit *t, void *_watch)
           if(t->evhooks->cancel_later)
             (*t->evhooks->cancel_later)(t->evdata, this);
           break;
+        case WATCH_SIGNAL:
+          if(t->evhooks->cancel_signal)
+            (*t->evhooks->cancel_signal)(t->evdata, this);
+          else
+            unwatch_signal(t, this);
+          break;
+        case WATCH_PROCESS:
+          if(t->evhooks->cancel_process)
+            (*t->evhooks->cancel_process)(t->evdata, this);
+          break;
 
-        default:
+        case WATCH_NONE:
           ;
       }
 
@@ -582,39 +849,6 @@ void tickit_evloop_set_watch_data_int(TickitWatch *watch, int data)
   watch->evdata.i = data;
 }
 
-static void invoke_watch(TickitWatch *watch, TickitEventFlags flags, void *info)
-{
-  (*watch->fn)(watch->t, flags, info, watch->user);
-
-  /* Remove oneshot watches from the list */
-  TickitWatch **prevp;
-  switch(watch->type) {
-    case WATCH_NONE:
-    case WATCH_IO:
-      return;
-
-    case WATCH_TIMER:
-      prevp = &watch->t->timers;
-      break;
-
-    case WATCH_LATER:
-      prevp = &watch->t->laters;
-      break;
-  }
-
-  while(*prevp) {
-    if(*prevp == watch) {
-      *prevp = watch->next;
-      watch->next = NULL;
-      watch->type = WATCH_NONE;
-      free(watch);
-      return;
-    }
-
-    prevp = &(*prevp)->next;
-  }
-}
-
 void tickit_evloop_invoke_watch(TickitWatch *watch, TickitEventFlags flags)
 {
   invoke_watch(watch, flags, NULL);
@@ -628,12 +862,28 @@ void tickit_evloop_invoke_iowatch(TickitWatch *watch, TickitEventFlags flags, Ti
   });
 }
 
+void tickit_evloop_invoke_processwatch(TickitWatch *watch, TickitEventFlags flags, int wstatus)
+{
+  invoke_watch(watch, flags, &(TickitProcessWatchInfo){
+    .pid     = watch->process.pid,
+    .wstatus = wstatus,
+  });
+}
+
+void tickit_evloop_invoke_sigwatches(Tickit *t, int signum)
+{
+  TickitWatch *this;
+  for(this = t->signals; this; this = this->next) {
+    if(this->signal.signum == signum)
+      (*this->fn)(this->t, TICKIT_EV_FIRE, NULL, this->user);
+  }
+}
+
 void tickit_evloop_sigwinch(Tickit *t)
 {
-  if(!t->term)
-    return;
-
-  tickit_term_refresh_size(t->term);
+  /* No longer need to do anything, as on_sigwinch() has already handled it
+   * But this function needs to remain for ABI reasons
+   */
 }
 
 const char *tickit_ctlname(TickitCtl ctl)
